<?php

namespace Drupal\forms_steps\Service;

use Drupal\Core\Entity\EntityTypeManagerInterface;
use Drupal\entity_embed\Exception\EntityNotFoundException;
use Drupal\Core\Entity\Entity;
use Drupal\forms_steps\Entity\Workflow;

/**
 * Class WorkflowManager.
 *
 * @package Drupal\forms_steps\Service
 */
class WorkflowManager {
  /**
   * EntityTypeManager.
   *
   * @var \Drupal\Core\Entity\EntityTypeManagerInterface
   */
  private $entityTypeManager;

  /**
   * WorkflowManager constructor.
   *
   * @param \Drupal\Core\Entity\EntityTypeManagerInterface $entity_type_manager
   *   EntityTypeManager provided by dependency injection.
   */
  public function __construct(
    EntityTypeManagerInterface $entity_type_manager
  ) {
    $this->entityTypeManager = $entity_type_manager;
  }

  /**
   * Returns the first workflow entry of the provided entity.
   *
   * @param \Drupal\Core\Entity\Entity $entity
   *   Entity instance to get the workflow from.
   *
   * @return \Drupal\forms_steps\Entity\Workflow|null
   *   Returns the workflow if found, null otherwise.
   *
   * @throws \Drupal\Component\Plugin\Exception\InvalidPluginDefinitionException
   * @throws \Drupal\Component\Plugin\Exception\PluginNotFoundException
   */
  public function getWorkflowByEntity(Entity $entity) : ?Workflow {
    $workflow = NULL;

    // We load all the workflow of that entity type & bundle.
    $workflows = \Drupal::entityTypeManager()
      ->getStorage(Workflow::ENTITY_TYPE)
      ->loadByProperties(
        [
          'entity_type' => $entity->getEntityTypeId(),
          'bundle' => $entity->bundle(),
          'entity_id' => $entity->id(),
        ]
      );

    // Only returning the first one.
    if (!empty($workflows)) {
      $workflow = reset($workflows);
    }

    return $workflow;
  }

  /**
   * Workflow info storage on hook_entity_presave().
   *
   * @param \Drupal\Core\Entity\Entity $entity
   *   The entity that is going to be saved.
   *
   * @throws \Drupal\Component\Plugin\Exception\InvalidPluginDefinitionException
   * @throws \Drupal\Component\Plugin\Exception\PluginNotFoundException
   * @throws \Drupal\Core\Entity\EntityStorageException
   * @throws \Drupal\entity_embed\Exception\EntityNotFoundException
   */
  public function entityPreSave(Entity $entity) {
    if ($entity->isNew()) {
      return;
    }

    $currentRoute = \Drupal::routeMatch()->getRouteName();
    if (
      preg_match('/^forms_steps\./', $currentRoute)
      && strcmp($entity->getEntityTypeId(), Workflow::ENTITY_TYPE) != 0
    ) {

      /** @var \Drupal\forms_steps\Service\FormsStepsManager $formsStepsManager */
      $formsStepsManager = \Drupal::service('forms_steps.manager');
      /** @var \Drupal\forms_steps\Step $step */
      $step = $formsStepsManager->getStepByRoute($currentRoute);

      if (is_null($step)) {
        throw new EntityNotFoundException(t('Unable to find Forms Steps step entity.'));
      }

      // We check if we are currently managing the right entity and not an
      // entity reference.
      if (strcmp($step->entityType(), $entity->getEntityTypeId()) != 0 ||
        strcmp($step->entityBundle(), $entity->bundle()) != 0) {
        return;
      }

      $instanceId = \Drupal::routeMatch()->getParameter('instance_id');
      $workflows = [];

      if (!empty($instanceId)) {
        try {
          $workflows = \Drupal::entityTypeManager()
            ->getStorage(Workflow::ENTITY_TYPE)
            ->loadByProperties(
              [
                'instance_id' => $instanceId,
                'entity_type' => $entity->getEntityTypeId(),
                'bundle' => $entity->bundle(),
                'step' => $step->id(),
                'entity_id' => $entity->id(),
                'form_mode' => $step->formMode(),
                'forms_steps' => $step->formsSteps()->id(),
              ]
            );
        }
        catch (\Exception $ex) {
          // Nothing to do.
        }
      }
      else {
        $instanceId = \Drupal::service('uuid')->generate();
      }

      if (empty($workflows)) {
        $workflow = \Drupal::entityTypeManager()
          ->getStorage(Workflow::ENTITY_TYPE)
          ->create([
            'instance_id' => $instanceId,
            'entity_type' => $entity->getEntityTypeId(),
            'bundle' => $entity->bundle(),
            'step' => $step->id(),
            'entity_id' => $entity->id(),
            'form_mode' => $step->formMode(),
            'forms_steps' => $step->formsSteps()->id(),
          ]);

        $workflow->save();
      }
    }
  }

  /**
   * Workflow info storage on hook_entity_insert().
   *
   * @param \Drupal\Core\Entity\Entity $entity
   *   The entity that is going to be inserted.
   *
   * @throws \Drupal\Component\Plugin\Exception\InvalidPluginDefinitionException
   * @throws \Drupal\Component\Plugin\Exception\PluginNotFoundException
   * @throws \Drupal\Core\Entity\EntityStorageException
   * @throws \Drupal\entity_embed\Exception\EntityNotFoundException
   */
  public function entityInsert(Entity $entity) {
    $currentRoute = \Drupal::routeMatch()->getRouteName();
    if (
      preg_match('/^forms_steps\./', $currentRoute)
      && strcmp($entity->getEntityTypeId(), Workflow::ENTITY_TYPE) != 0
    ) {

      /** @var \Drupal\forms_steps\Service\FormsStepsManager $formsStepsManager */
      $formsStepsManager = \Drupal::service('forms_steps.manager');
      /** @var \Drupal\forms_steps\Step $step */
      $step = $formsStepsManager->getStepByRoute($currentRoute);

      if (is_null($step)) {
        throw new EntityNotFoundException(t('Unable to find Forms Steps step entity.'));
      }

      // We check if we are currently managing the right entity and not an
      // entity reference.
      if (strcmp($step->entityType(), $entity->getEntityTypeId()) != 0 ||
        strcmp($step->entityBundle(), $entity->bundle()) != 0) {
        return;
      }

      $instanceId = \Drupal::routeMatch()->getParameter('instance_id');
      $workflows = [];

      if (!empty($instanceId)) {
        try {
          $workflows = \Drupal::entityTypeManager()
            ->getStorage(Workflow::ENTITY_TYPE)
            ->loadByProperties(
              [
                'instance_id' => $instanceId,
                'entity_type' => $entity->getEntityTypeId(),
                'bundle' => $entity->bundle(),
                'step' => $step->id(),
                'entity_id' => $entity->id(),
                'form_mode' => $step->formMode(),
                'forms_steps' => $step->formsSteps()->id(),
              ]
            );
        }
        catch (\Exception $ex) {
          // Nothing to do.
        }
      }
      else {
        $instanceId = \Drupal::service('uuid')->generate();
      }

      if (empty($workflows)) {
        $workflow = \Drupal::entityTypeManager()
          ->getStorage(Workflow::ENTITY_TYPE)
          ->create([
            'instance_id' => $instanceId,
            'entity_type' => $entity->getEntityTypeId(),
            'bundle' => $entity->bundle(),
            'step' => $step->id(),
            'entity_id' => $entity->id(),
            'form_mode' => $step->formMode(),
            'forms_steps' => $step->formsSteps()->id(),
          ]);

        $workflow->save();
      }
    }
  }

}
